@@ -2,8 +2,8 @@
 """SPF (Sender Policy Framework) implementation.
 
 Copyright (c) 2003, Terence Way
-Portions Copyright (c) 2004,2005,2006 Stuart Gathman <stuart@bmsi.com>
-Portions Copyright (c) 2005,2006 Scott Kitterman <scott@kitterman.com>
+Portions Copyright (c) 2004,2005,2006,2007 Stuart Gathman <stuart@bmsi.com>
+Portions Copyright (c) 2005,2006,2007 Scott Kitterman <scott@kitterman.com>
 This module is free software, and you may redistribute it and/or modify
 it under the same terms as Python itself, so long as this copyright message
 and disclaimer are retained in their original form.
@@ -28,45 +28,40 @@ For news, bugfixes, etc. visit the home page for this implementation at
     http://www.wayforward.net/spf/
 """
 
-# CVS Commits since last release (2.0.1):
+# CVS Commits since last release (2.0.2):
 # $Log: spf.py,v $
-# Revision 1.108.2.6  2006/12/30 17:12:50  customdesigned
-# Merge fixes from CVS HEAD.
+# Revision 1.108.2.13  2007/01/15 19:14:27  customdesigned
+# Permerror for more than one exp= or redirect=
 #
-# Revision 1.108.2.5  2006/12/24 19:10:38  kitterma
-# Move spf.py changelog to CHANGELOG. Move spf.py cvs commits from previous
-# releases to py_spfchangelog.txt. Update README to describe provided scripts.
-# Add to README discussion of spf module interface.
+# Revision 1.108.2.12  2007/01/13 18:45:33  customdesigned
+# Record matching mechanism.
 #
-# Revision 1.108.2.4  2006/12/23 06:35:37  customdesigned
-# Fully quote structured values in Received-SPF.
+# Revision 1.108.2.11  2007/01/13 18:21:41  customdesigned
+# Test for RFC4408 6.2/4, and fix spf.py to comply.
 #
-# Revision 1.108.2.3  2006/12/23 04:44:05  customdesigned
-# Fix key-value quoting in get_header.
+# Revision 1.123  2007/01/11 18:49:37  customdesigned
+# Add mechanism to Received-SPF header.
 #
-# Revision 1.121  2006/12/30 17:01:52  customdesigned
-# Missed a spot for new result names.
+# Revision 1.122  2007/01/11 18:25:54  customdesigned
+# Record matching mechanism.
 #
-# Revision 1.120  2006/12/28 04:54:21  customdesigned
-# Skip optional trailing ";" in Received-SPF
+# Revision 1.108.2.10  2007/01/13 00:46:35  kitterma
+# Update copyright statements for new year.
 #
-# Revision 1.118  2006/12/28 04:04:27  customdesigned
-# Optimize get_header to remove useless key-value pairs.
+# Revision 1.108.2.9  2007/01/12 22:14:56  kitterma
+# Change DNS queries to only check Type SPF in Harsh mode
 #
-# Revision 1.117  2006/12/23 06:31:16  customdesigned
-# Fully quote values in key-value pairs.
+# Revision 1.108.2.8  2007/01/06 22:58:21  kitterma
+# Update changelogs and version to reflect 2.0.2 released and 2.0.3 started.
 #
-# Revision 1.108.2.2  2006/12/22 20:27:24  customdesigned
-# Index error reporting non-mech permerror.
-#
-# Revision 1.108.2.1  2006/12/22 04:59:40  customdesigned
-# Merge comma heuristic.
+# Revision 1.108.2.7  2007/01/06 21:03:15  customdesigned
+# Tested spf.py in python2.2.
 #
 # See pyspf_changelog.txt for earlier CVS commits.
 
 __author__ = "Terence Way"
 __email__ = "terry@wayforward.net"
-__version__ = "2.0.2: December 23, 2006"
+__version__ = "2.0.3: January 12, 2007"
 MODULE = 'spf'
 
 USAGE = """To check an incoming mail request:
@@ -461,6 +456,8 @@ class query(object):
         # will continue processing.  However, the exception
         # that strict processing would raise is saved here
         self.perm_error = None
+	self.mechanism = None
+	self.options = {}
 
         try:
             self.lookups = 0
@@ -703,6 +700,7 @@ class query(object):
         default = 'neutral'
         mechs = []
 
+	modifiers = []
         # Look for modifiers
         #
         for mech in spf:
@@ -711,10 +709,17 @@ class query(object):
                 mechs.append(self.validate_mechanism(mech))
                 continue
 
+	    mod,arg = m
+	    if mod in modifiers:
+	    	if mod == 'redirect':
+		    raise PermError('redirect= MUST appear at most once',mech)
+		self.note_error('%s= MUST appear at most once'%mod,mech)
+		# just use last one in lax mode
+	    modifiers.append(mod)
             if m[0] == 'exp':
 	        # always fetch explanation to check permerrors
 	        exp = self.get_explanation(m[1])
-	        if not recursion:
+	        if exp and not recursion:
 		    # only set explanation in base recursion level
 		    self.set_explanation(exp)
             elif m[0] == 'redirect':
@@ -724,6 +729,10 @@ class query(object):
 		arg = self.expand(m[1])
                 # default=- is the same as default=fail
                 default = RESULTS.get(arg, default)
+	    elif mod == 'op':
+	    	if not recursion:
+		    for v in arg.split('.'):
+		        if v: self.options[v] = True
 	    else:
 		# spf rfc: 3.6 Unrecognized Mechanisms and Modifiers
 		self.expand(m[1])	# syntax error on invalid macro
@@ -796,11 +805,16 @@ class query(object):
                 if not redirect_record:
                     raise PermError('redirect domain has no SPF record',
                         redirect)
-                self.exps = dict(self.defexps)
+		# forget modifiers on redirect
+		if not recursion:
+		  self.exps = dict(self.defexps)
+		  self.options = {}
                 return self.check1(redirect_record, redirect, recursion)
-            else:
-                result = default
+	    result = default
+	    mech = None
 
+	if not recursion:	# record matching mechanism at base level
+	    self.mechanism = mech
         if result == 'fail':
             return (result, 550, exps[result])
         else:
@@ -816,10 +830,14 @@ class query(object):
     def get_explanation(self, spec):
         """Expand an explanation."""
         if spec:
-            txt = ''.join(self.dns_txt(self.expand(spec)))
-            return self.expand(txt, stripdot=False)
-        else:
-            return 'explanation : Required option is missing'
+            a = self.dns_txt(self.expand(spec))
+	    if len(a) == 1:
+	    	try:
+		    return self.expand(a[0], stripdot=False)
+		except PermError:
+		    # RFC4408 6.2/4 syntax errors cause exp= to be ignored
+		    pass
+	return None
 
     def expand(self, str, stripdot=True): # macros='slodipvh'
         """Do SPF RFC macro expansion.
@@ -975,21 +993,22 @@ class query(object):
         if len(a) == 1 and self.strict < 2:
             return a[0]               
         # check official SPF type first when it becomes more popular
-        try:
-            b = [t for t in self.dns_99(domain) if RE_SPF.match(t)]
-        except TempError,x:
-            # some braindead DNS servers hang on type 99 query
-            if self.strict > 1: raise TempError(x)
-            b = []
-
-        if len(b) > 1:
-            raise PermError('Two or more type SPF spf records found.')
-        if len(b) == 1:
-            if self.strict > 1 and len(a) == 1 and a[0] != b[0]:
-            #Changed from permerror to warning based on RFC 4408 Auth 48 change
-                raise AmbiguityWarning(
+        if self.strict > 1:
+            #Only check for Type SPF in harsh mode until it is more popular.
+            try:
+                b = [t for t in self.dns_99(domain) if RE_SPF.match(t)]
+            except TempError,x:
+                # some braindead DNS servers hang on type 99 query
+                if self.strict > 1: raise TempError(x)
+                b = []
+            if len(b) > 1:
+                raise PermError('Two or more type SPF spf records found.')
+            if len(b) == 1:
+                if self.strict > 1 and len(a) == 1 and a[0] != b[0]:
+                #Changed from permerror to warning based on RFC 4408 Auth 48 change
+                    raise AmbiguityWarning(
 'v=spf1 records of both type TXT and SPF (type 99) present, but not identical')
-            return b[0]
+                return b[0]
         if len(a) == 1:
             return a[0]    # return TXT if SPF wasn't found
         if DELEGATE:    # use local record if neither found
@@ -1156,8 +1175,10 @@ class query(object):
 	else:
 	    tag = res
 	    problem = None
+	mechanism = quote_value(self.mechanism)
 	res = ['%s (%s: %s)' % (tag,receiver,self.get_header_comment(res))]
-	for k in ('client_ip','envelope_from','helo','receiver','problem'):
+	for k in ('client_ip','envelope_from','helo','receiver',
+	  'problem','mechanism'):
 	    v = locals()[k]
 	    if v: res.append('%s=%s;'%(k,v))
 	res.append('identity=%s'%self.ident)
@@ -1238,8 +1259,13 @@ def quote_value(s):
 
     >>> quote_value('abc..def')
     '"abc..def"'
+
+    >>> quote_value('')
+    '""'
+
+    >>> quote_value(None)
     """
-    if RE_DOT_ATOM.match(s):
+    if s is None or RE_DOT_ATOM.match(s):
       return s
     return '"' + s.replace('\\',r'\\').replace('"',r'\"') + '"'
 
@@ -1362,7 +1388,7 @@ def bin2long6(str):
     h, l = struct.unpack("!QQ", str)
     return h << 64 | l
 
-if socket.has_ipv6:
+if hasattr(socket,'has_ipv6') and socket.has_ipv6:
     def inet_ntop(s):
         return socket.inet_ntop(socket.AF_INET6,s)
     def inet_pton(s):
@@ -1569,7 +1595,7 @@ if __name__ == '__main__':
         i, s, h = sys.argv[2:]
         q = query(i=i, s=s, h=h, receiver=socket.gethostname(),
             strict=False)
-        print q.check(sys.argv[1])
+        print q.check(sys.argv[1]),q.mechanism
         if q.perm_error and q.perm_error.ext:
             print q.perm_error.ext
     else:
